<?php

interface Applyable{
	// l'interface Applyable spécifie qu'une méthode 
	// "apply" doit exister et qu'elle prend en argument un objet
	// appelable (une fonction).
	function apply(callable $f);
}

/* Une classe abstraite : 
 * - elle ne peut pas être instanciée directement
 * - une autre classe peut en hériter, abstraite ou pas
 * - on peut définir une implémentation d'une fonction à
 * l'intérieur, ou bien déclarer des méthodes sans en donner 
 * l'implémentation en utilisant le mot-clef abstract.
 */
abstract class Liste{
	// renvoie le prochain élément de la liste
	abstract function next();
	// renvoie si la liste est vide
	abstract function isEmpty();
	// dit si l'élément courant (celui renvoyé par next) est
	// la fin de la liste (I.E. s'il existe).
	abstract function isEnd();


	// affiche la liste
	function printList(){
		$this->rewind();
		while(!$this->isEnd()){
			$el = $this->next();
			echo "$el, ";
		}
	}
}

// Une classe pour les maillons de la liste chaînée
class Element {
		public $next;
		public $value;
}

/* La classe ListeChainee hérite de Liste et respecte
 * l'interface Applyable en implémentant la méthode map
 */
class ListeChainee extends Liste implements Applyable{
	private $first; // référence vers la tête de la liste
	private $cur; // curseur pour next() et prev()


	// insère les éléments du tableau dans la liste
	// un à un et met le curseur au début.
	function __construct($array){
		foreach($array as $k => $v){
			$this->insert($v);
		}
		$this->rewind();
	}

	// remet le curseur au début
	function rewind(){
		$this->cur = $this->first;
	}

	// insère un élément
	function insert($el){
		$maillon = new Element();
		$maillon->next = $this->first;
		$this->first = $maillon;
	  $maillon->value = $el;	
	}

	// renvoie le prochain élément s'il existe.
	function next(){
		if(!$this->isEnd()){
			$val = $this->cur->value;
			$this->cur = $this->cur->next;
			return $val;
		} else return null;
	}
	
	// la liste est-elle vide ?
	function isEmpty(){
		return $this->first===null;
	}

	// le curseur est-il à la fin ?
	function isEnd(){
		return $this->cur===null;
	}

	// on applique f à tous les éléments de la liste
	function apply($f)
	{
		$this->rewind();
		while($this->isEnd()===false){
			$e = $this->next();
			$f($e);
		}
	}
}

// convertir prend deux arguments :
// * un tableau (forcé avec le type hint "array")
// * une Liste (forcé avec le type hint "Liste")
function convertir(array $array, Liste $liste){
	foreach($array as $v){
		$liste->insert($v);
	}
}

$liste = new ListeChainee([10,11]);
convertir([1,2,3,4],$liste);

$liste->apply(function($e){echo "$e, ";});
